@@ -34,6 +34,10 @@ func (arg *Argument) Value() Literal {
 	return arg.value
 }
 
+func (arg *Argument) IsSensitive() bool {
+	return arg.isSensitive
+}
+
 func (arg *Argument) WithValue(value Literal) *Argument {
 	return NewArgument(arg.Name(), value, arg.isSensitive)
 }
