@@ -27,6 +27,10 @@ func (m *Module) ID() *ID {
 	return m.id
 }
 
+func (m *Module) Name() string {
+	return m.pb.Name
+}
+
 func (m *Module) gatherCalls(callsByDigest map[string]*callpbv1.Call) {
 	if m == nil {
 		return
