@@ -197,11 +197,9 @@ func (db *DB) WalkSpans(opts FrontendOpts, spans iter.Seq[*Span], f func(*TraceT
 			lastCall = tree
 		}
 
-		children, revealed := span.ChildOrRevealedSpans(opts)
+		tree.RevealedChildren = len(span.RevealedSpans.Order) > 0
 
-		tree.RevealedChildren = revealed
-
-		for _, child := range children.Order {
+		for _, child := range span.ChildSpans.Order {
 			walk(child, tree)
 		}
 
@@ -244,7 +242,7 @@ func (lv *RowsView) Rows(opts FrontendOpts) *Rows {
 			Depth:                   depth,
 			IsRunningOrChildRunning: tree.IsRunningOrChildRunning,
 
-			HasChildren: len(tree.Children) > 0,
+			HasChildren: tree.hasVisibleChildren(opts),
 			Expanded:    tree.IsExpanded(opts),
 		}
 		if len(rows.Order) > 0 {
@@ -256,13 +254,29 @@ func (lv *RowsView) Rows(opts FrontendOpts) *Rows {
 		rows.BySpan[tree.Span.ID] = row
 		if row.Expanded {
 			var lastChild *TraceRow
-			for _, child := range tree.Children {
-				childRow := walk(child, row, depth+1)
-				if lastChild != nil {
-					childRow.Previous = lastChild
-					lastChild.Next = childRow
+
+			if tree.shouldShowRevealedSpans(opts) {
+				// Show revealed spans directly, finding their TraceTrees
+				for _, revealedSpan := range tree.Span.RevealedSpans.Order {
+					if revealedTree, ok := lv.BySpan[revealedSpan.ID]; ok {
+						childRow := walk(revealedTree, row, depth+1)
+						if lastChild != nil {
+							childRow.Previous = lastChild
+							lastChild.Next = childRow
+						}
+						lastChild = childRow
+					}
+				}
+			} else {
+				// Show direct children
+				for _, child := range tree.Children {
+					childRow := walk(child, row, depth+1)
+					if lastChild != nil {
+						childRow.Previous = lastChild
+						lastChild.Next = childRow
+					}
+					lastChild = childRow
 				}
-				lastChild = childRow
 			}
 			row.ShowingChildren = row.HasChildren
 		}
@@ -280,22 +294,37 @@ func (lv *RowsView) Rows(opts FrontendOpts) *Rows {
 	return rows
 }
 
+func (row *TraceTree) shouldShowRevealedSpans(opts FrontendOpts) bool {
+	verbosity := opts.Verbosity
+	if v, ok := opts.SpanVerbosity[row.Span.ID]; ok {
+		verbosity = v
+	}
+	return row.RevealedChildren && !opts.RevealNoisySpans && verbosity < ShowSpammyVerbosity
+}
+
+func (row *TraceTree) hasVisibleChildren(opts FrontendOpts) bool {
+	if row.shouldShowRevealedSpans(opts) {
+		return len(row.Span.RevealedSpans.Order) > 0
+	} else {
+		return len(row.Children) > 0
+	}
+}
+
 func (row *TraceTree) IsExpanded(opts FrontendOpts) bool {
 	expanded, toggled := opts.SpanExpanded[row.Span.ID]
 	if toggled {
 		return expanded
 	}
 
-	autoExpand := row.Depth() < 1 &&
-		(row.RevealedChildren || row.IsRunningOrChildRunning)
+	autoExpand := row.Depth() < 1 && row.IsRunningOrChildRunning
 
-	alwaysExpand := row.Span.IsFailedOrCausedFailure() ||
-		row.Span.IsCanceled() ||
-		opts.Verbosity >= ExpandCompletedVerbosity
+	alwaysExpand := row.Span.IsCanceled() ||
+		opts.Verbosity >= ExpandCompletedVerbosity ||
+		opts.ExpandCompleted
 
 	// never expand tool calls by default, tends to show a bunch of guts that
 	// distracts from the overall history
-	neverExpand := row.Span.LLMTool != ""
+	neverExpand := row.Span.LLMTool != "" || row.Span.RollUpLogs || row.Span.RollUpSpans
 
 	return (autoExpand || alwaysExpand) && !neverExpand
 }
